/*
** Module   : EnvironmentOps.java
** Abstract : Progress 4GL compatible environment related methods
**
** Copyright (c) 2005-2011, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 GES 20050628   @21564 Created initial version, supporting common
**                           global variables that are called built-in
**                           functions in Progress but which take no
**                           parameters, no empty parens. These also
**                           are read-only (they cannot be assigned).
** 002 GES 20050729   @21921 Added getSourceName() and modified the
**                           default propath.
** 003 ECF 20051214   @23753 Removed getCurrentDatabaseName(). This was a
**                           placeholder for P2J's DBNAME replacement
**                           implementation. This is now provided in the
**                           ConnectionManager class instead.
** 004 GES 20060123   @24034 Move to a directory based implementation for
**                           the Progress compatibility global vars.
** 005 GES 20060125   @24071 Shift to SourceNameMapper backed stack
**                           processing.
** 006 EVL 20060713   @28035 Changing the approach to get the source file
**                           name from index of the Progress procedure on
**                           the stack.
** 007 GES 20070321   @32506 Added support for assigning the PROPATH.
** 008 GES 20071214   @36369 Fix for NPE in getSourceName().
** 009 GES 20080123   @36902 Better propath and separator defaults.
** 010 GES 20080321   @37628 Added SESSION:BATCH-MODE support.
** 011 GES 20080414   @38014 Added SESSION:CHARSET support.
** 012 GES 20080924   @39940 Mask off inner classes from program-name
**                           processing.
** 013 GES 20090512   @42147 Minor optimization for program-name.
** 014 ECF 20101005          Added getTempDirectory method.
** 015 GES 20110713          Changed default 4GL version string to "10.2B".
** 016 CS  20121104          Implement PROPATH assignment.
*/

package com.goldencode.p2j.util;

import java.io.File;
import java.util.*;
import com.goldencode.p2j.security.ContextLocal;

/**
 * Progress compatible environment-related static utility methods.
 * <p>
 * This supports some common global variables that are called built-in
 * functions in Progress but which take no parameters, not even empty
 * parenthesis. These variables are read-only (they cannot be assigned),
 * except for <code>PROPATH</code> which can be modified as it can be in
 * Progress as well.
 * <p>
 * The following is the list of supported variables:
 * <p>
 * <pre>
 * Progress Variable                    Java Equivalent
 * ---------------------                ----------------------
 * current-language                     {@link #getCurrentLanguage}
 * dataservers                          {@link #getDataServerList}
 * gateways                             {@link #getDataServerList}
 * opsys                                {@link #getOperatingSystem}
 * progress                             {@link #getRuntimeType}
 * promsgs                              {@link #getMessageSource}
 * propath                              {@link #getSearchPath}
 * proversion                           {@link #getVersion}
 * </pre>
 * <p>
 * <pre>
 * Progress Function                    Java Equivalent
 * ---------------------                ----------------------
 * program-name                         {@link #getSourceName}
 * </pre>
 * <p>
* <pre>
 * System Handle:Attribute              Java Equivalent
 * -------------------------            ----------------------
 * session:batch-mode                   {@link #isBatchMode}       (get)
 * session:charset                      {@link #getCharset}        (get)
 * session:temp-directory               {@link #getTempDirectory}  (get)
 * </pre>
 */
public class EnvironmentOps
{
   /** Progress OPSYS string for Unix systems. */
   public static final String OS_UNIX = "UNIX";
   
   /** Progress OPSYS string for Windows systems. */
   public static final String OS_WIN = "WIN32";
   
   /** 
    * Maps known Java os.name system property entries to a given Progress 
    * OPSYS string.  Entries should be inserted in lowercase and all keys
    * should be lowercased before reading from the map.
    */
   public static final Map osMap = new HashMap();
   
   static
   {
      // unix variants
      osMap.put("linux",    OS_UNIX);
      osMap.put("aix",      OS_UNIX);
      osMap.put("solaris",  OS_UNIX);
      osMap.put("sunos",    OS_UNIX);
      osMap.put("hp-ux",    OS_UNIX);
      osMap.put("freebsd",  OS_UNIX);
      osMap.put("mac os x", OS_UNIX);
      
      // windows variants
      osMap.put("windows xp",   OS_WIN);
      osMap.put("windows 2000", OS_WIN);
      osMap.put("windows nt",   OS_WIN);
      osMap.put("windows me",   OS_WIN);
      osMap.put("windows 98",   OS_WIN);
      osMap.put("windows 95",   OS_WIN);
   }
   
   /** Stores context-local state variables. */
   private static ContextContainer work = new ContextContainer(); 
   
   /**
    * Gets the name of the current language being used for the Progress
    * compatible environment.
    * <p>
    * The value returned may have been found via a search algorithm that
    * is account (user or process) specific or group specific
    * within the current server or a global default for all servers. 
    * <p>
    * The implementation iteratively looks up the directory node under:
    * /server/&lt;serverID&gt;/runtime/&lt;account_or_group&gt;/currentLanguage
    * <p>
    * If no user/process or group nodes are present, then this is checked:
    * /server/&lt;serverID&gt;/runtime/default/currentLanguage
    * <p>
    * If no /server/&lt;serverID&gt;/runtime node exists, this is checked
    * (it is the global default area for all servers):
    * /server/default/runtime/&lt;account_or_group&gt;/currentLanguage
    * <p>
    * Finally, if no user/process or group nodes are present in the global
    * default area, then this is checked:
    * /server/default/runtime/default/currentLanguage
    * <p>
    * If no value is found via this lookup, then the default value of
    * "?" will be returned.
    *
    * @return   The name of the current language.
    */
   public static character getCurrentLanguage()
   {
      String lang = Utils.getDirectoryNodeString(null,
                                                 "currentLanguage",
                                                 "?");
      return new character(lang);
   }
   
   /**
    * Gets the list of the data servers available in Progress compatible
    * environment.
    * <p>
    * The value returned may have been found via a search algorithm that
    * is account (user or process) specific or group specific
    * within the current server or a global default for all servers. 
    * <p>
    * The implementation iteratively looks up the directory node under:
    * /server/&lt;serverID&gt;/runtime/&lt;account_or_group&gt;/dataServers
    * <p>
    * If no user/process or group nodes are present, then this is checked:
    * /server/&lt;serverID&gt;/runtime/default/dataServers
    * <p>
    * If no /server/&lt;serverID&gt;/runtime node exists, this is checked
    * (it is the global default area for all servers):
    * /server/default/runtime/&lt;account_or_group&gt;/dataServers
    * <p>
    * Finally, if no user/process or group nodes are present in the global
    * default area, then this is checked:
    * /server/default/runtime/default/dataServers
    * <p>
    * If no value is found via this lookup, then the default value of
    * "PROGRESS,ORACLE,AS400,ODBC,MSS" will be returned.
    *
    * @return   The list of data servers.
    */
   public static character getDataServerList()
   {
      String def     = "PROGRESS,ORACLE,AS400,ODBC,MSS";
      String dserver = Utils.getDirectoryNodeString(null,
                                                    "dataServers",
                                                    def);
      return new character(dserver);
   }
   
   /**
    * Gets the name of the operating system being upon which the Progress
    * compatible environment is running.  The first attempt to resolve this
    * value will occur via a lookup in the directory.
    * <p>
    * The value returned may have been found via a search algorithm that
    * is account (user or process) specific or group specific
    * within the current server or a global default for all servers. 
    * <p>
    * The implementation iteratively looks up the directory node under:
    * /server/&lt;serverID&gt;/runtime/&lt;account_or_group&gt;/opsys/override
    * <p>
    * If no user/process or group nodes are present, then this is checked:
    * /server/&lt;serverID&gt;/runtime/default/opsys/override
    * <p>
    * If no /server/&lt;serverID&gt;/runtime node exists, this is checked
    * (it is the global default area for all servers):
    * /server/default/runtime/&lt;account_or_group&gt;/opsys/override
    * <p>
    * Finally, if no user/process or group nodes are present in the global
    * default area, then this is checked:
    * /server/default/runtime/default/opsys/override
    * <p>
    * If no value is found via this lookup, then the Java system property
    * "os.name" is queried.  This is then translated via an optional
    * mapping table in the directory (that same paths and precedence as
    * above but ending in "opsys/mapping/&lt;lowercased_system_propert_name&gt;.
    * <p>
    * If no mapping is found, then an internal mapping table in this class
    * is consulted.  This table maps the following:
    * <p>
    * <pre>
    * os.name          Progress OPSYS
    * ---------------- ---------------
    * "linux"          "UNIX"
    * "aix"            "UNIX"
    * "solaris"        "UNIX"
    * "sunos"          "UNIX"
    * "hp-ux"          "UNIX"
    * "freebsd"        "UNIX"
    * "mac os x"       "UNIX"
    * "windows xp"     "WIN32"
    * "windows 2000"   "WIN32"
    * "windows nt"     "WIN32"
    * "windows me"     "WIN32"
    * "windows 98"     "WIN32"
    * "windows 95"     "WIN32"
    * </pre>
    * <p>
    * If there is still no match, then any <code>os.name</code> with the
    * lowercase text "unix" will return "UNIX" and any <code>os.name</code>
    * with the lowercase text "windows" will return "WIN32". 
    * <p>
    * If there is still no match, the <code>os.name</code> returned by the
    * Java will be returned. 
    *
    * @return   The name of the current operating system.
    */
   public static character getOperatingSystem()
   {
      // look for a hard coded value
      String osname = Utils.getDirectoryNodeString(null,
                                                   "opsys/override",
                                                   "");
                                                      
      // if there is a hard coded operating system override, use it
      if (osname.length() == 0)
      {
         // query the JVM value (it will always we present)
         String fallback = System.getProperty("os.name");
         String sysprop  = fallback.toLowerCase();
         
         // check if there is a mapping in the directory for this exact
         // value
         osname = "opsys/mapping/" + sysprop;
         osname = Utils.getDirectoryNodeString(null, osname, "");
                                                   
         // use the mapping if there was one
         if (osname.length() == 0)
         {
            // attempt to translate it by using our predefined tables
            osname = (String) osMap.get(sysprop);
            
            // no match?
            if (osname == null)
            {
               // last try
               if (sysprop.indexOf("unix") != -1)
               {
                  osname = OS_UNIX;
               }
               else if (sysprop.indexOf("windows") != -1)
               {
                  osname = OS_WIN;
               }
               else
               {
                  // all else fails, return what Java specified
                  osname = fallback;
               }
            }
         }
      }
      
      return new character(osname);
   }
   
   /**
    * Gets the current runtime type being provided by the Progress
    * compatible environment.  The P2J environment only provides a
    * "Full" environment.
    *
    * @return   Always returns the string "Full".
    */
   public static character getRuntimeType()
   {
      return new character("Full");
   }
   
   /**
    * Gets the name of the directory where temporary files are stored.
    * Corresponds with with the <code>SESSION:TEMP-DIRECTORY</code> system
    * handle attribute.
    * 
    * @return  Name of temporary directory.
    */
   public static character getTempDirectory()
   {
      return new character(System.getProperty("java.io.tmpdir"));
   }
   
   /**
    * Gets the name of the current message source being used for the Progress
    * compatible environment.
    *
    * @return   Always returns the string "promsgs".
    */
   public static character getMessageSource()
   {
      return new character("promsgs");
   }
   
   /**
    * Sets the Progress search path on the TARGET P2J system which is
    * used for the Progress compatible environment.  This corresponds to the
    * assignment of the "PROPATH" variable in Progress and represents a
    * mechanism to override the default that is available from the directory.
    * <p>
    * The PROPATH is also used in the implementation of
    * {@link FileSystemOps#searchPath}.  The actual I/O operations when
    * executed via the Progress compatible interface (see
    * {@link FileSystemOps} and {@link Stream}) for a P2J system are executed
    * on the P2J client.  This is the TARGET system on which this path is
    * valid.
    * <p>
    * This must be distinguished from the original Progress SOURCE system's
    * PROPATH which must be maintained for certain runtime translations
    * of dynamically generated names (the algorithms of which were designed
    * with that path assumed) but which is not necessarily a valid path on 
    * the system on which this is running.  The original PROPATH can be
    * obtained via the {@link #getLegacySearchPath} method.
    *
    * @param    path
    *           The replacement path to Progress source files.
    */
   public static void setSearchPath(String path)
   {
      work.obtain().propath = path;
      if (path != null)
      {
         String[] pathList = splitPath(path);
         FileSystemOps.setSearchPath(pathList, getLegacyCaseSensitive()); 
      }
   }
   
   /**
    * Sets the Progress search path on the TARGET P2J system which is
    * used for the Progress compatible environment.  This corresponds to the
    * assignment of the "PROPATH" variable in Progress and represents a
    * mechanism to override the default that is available from the directory.
    * <p>
    * The PROPATH is also used in the implementation of
    * {@link FileSystemOps#searchPath}.  The actual I/O operations when
    * executed via the Progress compatible interface (see
    * {@link FileSystemOps} and {@link Stream}) for a P2J system are executed
    * on the P2J client.  This is the TARGET system on which this path is
    * valid.
    * <p>
    * This must be distinguished from the original Progress SOURCE system's
    * PROPATH which must be maintained for certain runtime translations
    * of dynamically generated names (the algorithms of which were designed
    * with that path assumed) but which is not necessarily a valid path on 
    * the system on which this is running.  The original PROPATH can be
    * obtained via the {@link #getLegacySearchPath} method.
    *
    * @param    path
    *           The replacement path to Progress source files.
    */
   public static void setSearchPath(character path)
   {
      setSearchPath(path.toStringMessage());
   }
   
   /**
    * Gets any context-specific override for the Progress search path which
    * was previously set via {@link #setSearchPath}.
    *
    * @return   The path override or <code>null</code> if no override was
    *           ever set.
    */
   public static String getSearchPathOverride()
   {
      return work.obtain().propath;
   }
   
   /**
    * Gets the Progress search path on the TARGET P2J system which is
    * used for the Progress compatible environment.  This corresponds to the
    * use of the "PROPATH" variable in Progress and is also used in the
    * implementation of {@link FileSystemOps#searchPath}.  The actual I/O
    * operations when executed via the Progress compatible interface (see
    * {@link FileSystemOps} and {@link Stream}) for a P2J system are executed
    * on the P2J client.  This is the TARGET system on which this path is
    * valid.
    * <p>
    * This must be distinguished from the original Progress SOURCE system's
    * PROPATH which must be maintained for certain runtime translations
    * of dynamically generated names (the algorithms of which were designed
    * with that path assumed) but which is not necessarily a valid path on 
    * the system on which this is running.  The original PROPATH can be
    * obtained via the {@link #getLegacySearchPath} method.
    * <p>
    * If a specific PROPATH has been assigned in this user's context, then
    * that "override" will be returned.  Otherwise, the directory will be
    * consulted for the stored version.
    * <p>
    * The value returned from the directory may have been found via a search
    * algorithm that is account (user or process) specific or group specific
    * within the current server or a global default for all servers. 
    * <p>
    * The implementation iteratively looks up the directory node under:
    * /server/&lt;serverID&gt;/runtime/&lt;account_or_group&gt;/searchpath
    * <p>
    * If no user/process or group nodes are present, then this is checked:
    * /server/&lt;serverID&gt;/runtime/default/searchpath
    * <p>
    * If no /server/&lt;serverID&gt;/runtime node exists, this is checked
    * (it is the global default area for all servers):
    * /server/default/runtime/&lt;account_or_group&gt;/searchpath
    * <p>
    * Finally, if no user/process or group nodes are present in the global
    * default area, then this is checked:
    * /server/default/runtime/default/searchpath
    * <p>
    * If no value is found via this lookup, then the default value of
    * ".:" will be returned.
    *
    * @return   The path to Progress source files.
    */
   public static character getSearchPath()
   {
      WorkArea wa      = work.obtain();
      String   propath = null;
      
      // is there a context-specific override?
      if (wa.propath != null)
      {
         // support the override
         propath = wa.propath;
      }
      else
      {
         // lookup in the directory and fallback to a default if none exists
         propath = Utils.getDirectoryNodeString(null,
                                                "searchpath",
                                                ".:");
      }
      
      return new character(propath);
   }
   
   /**
    * Gets the original Progress SOURCE system's PROPATH which must be
    * maintained for certain runtime translations of dynamically generated
    * names (the algorithms of which were designed with that path assumed)
    * but which is not necessarily a valid path on the system on which this
    * is running.
    * <p>
    * The value returned may have been found via a search algorithm that
    * is account (user or process) specific or group specific
    * within the current server or a global default for all servers. 
    * <p>
    * The implementation iteratively looks up the directory node under:
    * /server/&lt;serverID&gt;/runtime/&lt;account_or_group&gt;/file-system/propath
    * <p>
    * If no user/process or group nodes are present, then this is checked:
    * /server/&lt;serverID&gt;/runtime/default/file-system/propath
    * <p>
    * If no /server/&lt;serverID&gt;/runtime node exists, this is checked
    * (it is the global default area for all servers):
    * /server/default/runtime/&lt;account_or_group&gt;/file-system/propath
    * <p>
    * Finally, if no user/process or group nodes are present in the global
    * default area, then this is checked:
    * /server/default/runtime/default/file-system/propath
    * <p>
    * If no value is found via this lookup, then the default value of
    * "." will be returned.
    * <p>
    * This must be distinguished from Progress source file/search path on the
    * TARGET P2J system which is used for the Progress compatible
    * environment.  To obtain that value, use {@link #getSearchPath}.
    *
    * @return   The path to Progress source files.
    */
   public static String getLegacySearchPath()
   {
      return Utils.getDirectoryNodeString(null, "file-system/propath", ".");
   }
   
   /**
    * Gets the path separator character used to parse the Progress search
    * path (see {@link #getSearchPath}).
    * <p>
    * The value returned may have been found via a search algorithm that
    * is account (user or process) specific or group specific
    * within the current server or a global default for all servers. 
    * <p>
    * The implementation iteratively looks up the directory node under:
    * /server/&lt;serverID&gt;/runtime/&lt;account_or_group&gt;/path-separator
    * <p>
    * If no user/process or group nodes are present, then this is checked:
    * /server/&lt;serverID&gt;/runtime/default/path-separator
    * <p>
    * If no /server/&lt;serverID&gt;/runtime node exists, this is checked
    * (it is the global default area for all servers):
    * /server/default/runtime/&lt;account_or_group&gt;/path-separator
    * <p>
    * Finally, if no user/process or group nodes are present in the global
    * default area, then this is checked:
    * /server/default/runtime/default/path-separator
    * <p>
    * If no value is found via this lookup, then the default value of
    * ':' will be returned.
    *
    * @return   The path to Progress source files.
    */
   public static String getPathSeparator()
   {
      return Utils.getDirectoryNodeString(null, "path-separator", ":");
   }
   
   /**
    * Gets the original Progress SOURCE system's path separator which must be
    * maintained for certain runtime translations of dynamically generated
    * names (the algorithms of which were designed with that path assumed)
    * but which is not necessarily a valid path on the system on which this
    * is running.
    * <p>
    * The value returned may have been found via a search algorithm that
    * is account (user or process) specific or group specific
    * within the current server or a global default for all servers. 
    * <p>
    * The implementation iteratively looks up the directory node under:
    * /server/&lt;serverID&gt;/runtime/&lt;account_or_group&gt;/file-system/path-separator
    * <p>
    * If no user/process or group nodes are present, then this is checked:
    * /server/&lt;serverID&gt;/runtime/default/file-system/path-separator
    * <p>
    * If no /server/&lt;serverID&gt;/runtime node exists, this is checked
    * (it is the global default area for all servers):
    * /server/default/runtime/&lt;account_or_group&gt;/file-system/path-separator
    * <p>
    * Finally, if no user/process or group nodes are present in the global
    * default area, then this is checked:
    * /server/default/runtime/default/file-system/path-separator
    * <p>
    * If no value is found via this lookup, then the default value of
    * ":" will be returned.
    * <p>
    * This must be distinguished from the path separator on the (runtime)
    * TARGET P2J system which is used for the Progress compatible
    * environment.  To obtain that value, use {@link #getPathSeparator}.
    *
    * @return   The path to Progress source files.
    */
   public static String getLegacyPathSeparator()
   {
      return Utils.getDirectoryNodeString(null,
                                          "file-system/path-separator",
                                          ":");
   }
   
   /**
    * Gets the list of directories in the Progress source file path.  Each
    * element of the array is a directory in the source path.  This source
    * path is split based on the path separator used in this path
    * specification which may be different than that reported by the
    * system (<code>File.pathSeparator</code>.  The resulting array is
    * returned to the caller.
    * <p>
    * The {@link #getSearchPath} is used for the path specification and
    * the {@link #getPathSeparator} is used to split the path.
    *
    * @return   The array of directory names.
    */
   public static String[] getSearchPathList()
   {
      return splitPath(getSearchPath().getValue());
   }
   
   /**
    * Gets the original Progress SOURCE system's file separator which must be
    * maintained for certain runtime translations of dynamically generated
    * names (the algorithms of which were designed with that path assumed)
    * but which is not necessarily a valid path on the system on which this
    * is running.
    * <p>
    * The value returned may have been found via a search algorithm that
    * is account (user or process) specific or group specific
    * within the current server or a global default for all servers. 
    * <p>
    * The implementation iteratively looks up the directory node under:
    * /server/&lt;serverID&gt;/runtime/&lt;account_or_group&gt;/file-system/file-separator
    * <p>
    * If no user/process or group nodes are present, then this is checked:
    * /server/&lt;serverID&gt;/runtime/default/file-system/file-separator
    * <p>
    * If no /server/&lt;serverID&gt;/runtime node exists, this is checked
    * (it is the global default area for all servers):
    * /server/default/runtime/&lt;account_or_group&gt;/file-system/file-separator
    * <p>
    * Finally, if no user/process or group nodes are present in the global
    * default area, then this is checked:
    * /server/default/runtime/default/file-system/file-separator
    * <p>
    * If no value is found via this lookup, then the default value of
    * "/" will be returned.
    *
    * @return   The legacy system's file separator.
    */
   public static String getLegacyFileSeparator()
   {
      return Utils.getDirectoryNodeString(null,
                                          "file-system/file-separator",
                                          "/");
   }
   
   /**
    * Gets the original Progress SOURCE file system's case-sensitive setting
    * which must be maintained for certain runtime translations of dynamically
    * generated names (the algorithms of which were designed with that path
    * assumed) but which is not necessarily a valid path on the system on
    * which this is running.
    * <p>
    * The value returned may have been found via a search algorithm that
    * is account (user or process) specific or group specific
    * within the current server or a global default for all servers. 
    * <p>
    * The implementation iteratively looks up the directory node under:
    * /server/&lt;serverID&gt;/runtime/&lt;account_or_group&gt;/file-system/case-sensitive
    * <p>
    * If no user/process or group nodes are present, then this is checked:
    * /server/&lt;serverID&gt;/runtime/default/file-system/case-sensitive
    * <p>
    * If no /server/&lt;serverID&gt;/runtime node exists, this is checked
    * (it is the global default area for all servers):
    * /server/default/runtime/&lt;account_or_group&gt;/file-system/case-sensitive
    * <p>
    * Finally, if no user/process or group nodes are present in the global
    * default area, then this is checked:
    * /server/default/runtime/default/file-system/case-sensitive
    * <p>
    * If no value is found via this lookup, then the default value of
    * <code>true</code> will be returned.
    *
    * @return   <code>true</code> if the legacy file system was 
    *           case-sensitive.
    */
   public static boolean getLegacyCaseSensitive()
   {
      return Utils.getDirectoryNodeBoolean(null,
                                           "file-system/case-sensitive",
                                           true);
   }
   
   /**
    * Gets the current version string of the Progress compatible environment.
    * <p>
    * The value returned may have been found via a search algorithm that
    * is account (user or process) specific or group specific
    * within the current server or a global default for all servers. 
    * <p>
    * The implementation iteratively looks up the directory node under:
    * /server/&lt;serverID&gt;/runtime/&lt;account_or_group&gt;/version
    * <p>
    * If no user/process or group nodes are present, then this is checked:
    * /server/&lt;serverID&gt;/runtime/default/version
    * <p>
    * If no /server/&lt;serverID&gt;/runtime node exists, this is checked
    * (it is the global default area for all servers):
    * /server/default/runtime/&lt;account_or_group&gt;/version
    * <p>
    * Finally, if no user/process or group nodes are present in the global
    * default area, then this is checked:
    * /server/default/runtime/default/version
    * <p>
    * If no value is found via this lookup, then the default value of
    * "9.1C" will be returned.
    *
    * @return   The current version string.
    */
   public static character getVersion()
   {
      String version = Utils.getDirectoryNodeString(null,
                                                    "version",
                                                    "10.2B");
      return new character(version);
   }
   
   /**
    * Returns the Progress source file name that matches the program at the
    * given index in the stack trace if the given index position exists. The 
    * Progress source file name will be returned if possible using the
    * {@link SourceNameMapper#convertStackEntry}.  If it does not exist, then
    * the Java classname of the entry will be returned.
    * <p>
    * Limitations:
    * <p>
    * <ol>
    *    <li> Progress returns the same filename as is used on the RUN
    *         statement (so the name can change from invocation to
    *         invocation).  This version has a static string for the base
    *         name of the file (with relative pathing).
    *    <li> It is possible that due to code refactoring, code that calls 
    *         program-name might have dependencies on a source file name that
    *         is not in the stack trace of the Java version.
    * </ol>
    *
    * @param    index
    *           The 1-based index position in the call stack to be returned.
    *
    * @return   The Progress source file name (if the class has the proper
    *           member) or otherwise the Java class name, if the given index
    *           position exists.  If the index is out of bounds, then the
    *           <code>unknown value</code> is returned.
    */
   public static character getSourceName(integer index)
   {
      if (index.isUnknown())
      {
         return new character();
      }
      
      return getSourceName(index.intValue());
   }
   
   /**
    * Returns the Progress source file name that matches the program at the
    * given index in the stack trace if the given index position exists. The 
    * Progress source file name will be returned if possible using the
    * {@link SourceNameMapper#convertStackEntry}.  If it does not exist, then
    * the Java classname of the entry will be returned.
    * <p>
    * Limitations:
    * <p>
    * <ol>
    *    <li> Progress returns the same filename as is used on the RUN
    *         statement (so the name can change from invocation to
    *         invocation).  This version has a static string for the base
    *         name of the file (with relative pathing).
    *    <li> It is possible that due to code refactoring and/or execution of
    *         a runtime that has an arbitrarily different implementation than
    *         does Progress, code that calls program-name might have
    *         dependencies on a source file name that is not in the stack
    *         trace of the Java version.
    * </ol>
    *
    * @param    index
    *           The 1-based index position in the call stack to be returned.
    *
    * @return   The Progress source file name (if the class is properly
    *           mapped in the directory) or otherwise the Java class name,
    *           if the given index position exists.  If the index is out of
    *           bounds, then the <code>unknown value</code> is returned.
    */
   public static character getSourceName(int index)
   {
      // get a helper to process the stack
      Throwable stack = new Throwable();
      
      // obtain the stack trace
      stack.fillInStackTrace();
      StackTraceElement[] trace = stack.getStackTrace();
      
      String result = null;
      String method = null;
      
      // Note: although the index is 1-based and our array is 0-based, the
      // trace[0] is by definition always going to be defined by this
      // class (EnvironmentOps) and method (getSourceName), so where we
      // would normally convert the index into a 0-based form, we need to
      // leave it as is so that it indexes into the array properly.
      
      // if inside the array bounds, obtain the class name, note that the
      // index request might be very large, still within the Java stack
      // bounds BUT NOT have any valid backing 4GL entry that matches
      if (index > 0 && index < trace.length)
      {
         WorkArea wa      = work.obtain();
         String   pkgroot = null;
         
         // is there a context-specific override?
         if (wa.pkgroot != null)
         {
            // support the override
            pkgroot = wa.pkgroot;
         }
         else
         {
            // lookup in the directory and fallback to a default if none exists
            pkgroot = Utils.getDirectoryNodeString(null,
                                                   "file-system/pkgroot",
                                                   "");
         }
         
         int indexFound = 0;
         
         // the Java stack trace will have many additional entries that
         // don't exist in the 4GL (all kinds of runtime and J2SE entries
         // will be in the stack), so we "mask" off all the entries that
         // aren't application level entries by comparing with pkgroot
         for (int i = 0; i < trace.length; i++)
         {
            // Note: it is possible that getFileName() needs to be used
            //       instead of getClassName() however no real requirement
            //       is known at this time.
            String cname = trace[i].getClassName();
            
            // assume that anything in the pkgroot is an application class
            // and is not part of the runtime (or J2SE)
            boolean notRuntime = cname.startsWith(pkgroot);
            
            // assume that any inner classes can be ignored (compact mode
            // heavily uses inner classes for inner block processing but
            // these never correspond to external or internal procedure
            // names
            int     lastDot  = cname.lastIndexOf('.');
            boolean notInner = cname.indexOf('$', lastDot) == -1;
            
            if (notRuntime && notInner)
               indexFound++;
            
            if (indexFound == index)
            {
               result = cname;
               method = trace[i].getMethodName();
               break;
            }
         }
         
         // if result and method are both null then there was no matching
         // 4GL entry for that index position in the Java stack, just drop
         // out and return unknown
         
         if (result != null)
         {
            if (method != null && method.equals("execute"))
            {
               // this is an external procedure, no method level info is
               // needed
               method = null;
            }
            
            String override = SourceNameMapper.convertStackEntry(result,
                                                                 method);
            
            if (override != null)
            {
               result = override;
            }         
            else
            {
               // no mapping available!
               // TODO: log this
            }
         }
      }
      
      // out of bounds index?
      if (result == null)
      {
         // return the unknown value
         return new character();
      }
      
      return new character(result);
   }
   
   /**
    * Splits a path using the pathSeparator
    * @param path the path to split
    * @return The array of directory names
    */
   private static String[] splitPath(String path)
   {
      if (path != null)
      {
         return path.split(getPathSeparator(), -1);
      }
      else
      {
         return null;
      }
   }
   
   /**
    * Reports if the current context is running in batch mode.
    * <p>
    * The value returned may have been found via a search algorithm that
    * is account (user or process) specific or group specific
    * within the current server or a global default for all servers. 
    * <p>
    * The implementation iteratively looks up the directory node under:
    * /server/&lt;serverID&gt;/runtime/&lt;account_or_group&gt;/batchMode
    * <p>
    * If no user/process or group nodes are present, then this is checked:
    * /server/&lt;serverID&gt;/runtime/default/batchMode
    * <p>
    * If no /server/&lt;serverID&gt;/runtime node exists, this is checked
    * (it is the global default area for all servers):
    * /server/default/runtime/&lt;account_or_group&gt;/batchMode
    * <p>
    * Finally, if no user/process or group nodes are present in the global
    * default area, then this is checked:
    * /server/default/runtime/default/batchMode
    * <p>
    * If no value is found via this lookup, then the default value of
    * <code>false</code> will be returned.
    *
    * @return   <code>true</code> if the current context is in batch mode.
    */
   public static logical isBatchMode()
   {
      boolean batch = Utils.getDirectoryNodeBoolean(null,
                                                    "batchMode",
                                                    false);
      return new logical(batch);
   }
   
   /**
    * Reports the default character set name. Since the Progress default
    * charset is ISO-8859-1, when the JVM reports that UTF-8 is the default,
    * this method will return the string "ISO-8859-1" instead.
    * <p>
    * The assumption (possibly not always correct) is that UTF-8 is not a
    * valid Progress charset and instead, to provide a proper default, we
    * must override with a compatible charset. Note that if the default
    * charset name is not UTF-8, it is assumed that the charset has been
    * overriden already (at JVM startup).  In that case, the default is
    * returned unchanged.
    *
    * @return   A default charset name.
    */
   public static character getCharset()
   {
      return new character(Utils.getCharsetName());
   }
   
   /** 
    * Stores global data relating to the state of the current context.
    */
   private static class WorkArea
   {  
      /** Stores any context-specific override to the <code>PROPATH</code>. */
      private String propath = null;
      
      /** Cache the context-specific package prefix. */
      private String pkgroot = null;
   }
   
   /**
    * Simple container that stores and returns a context-local instance of
    * the global work area.
    */
   private static class ContextContainer
   extends ContextLocal
   {
      /**
       * Obtains the context-local instance of the contained global work
       * area.
       *
       * @return   The work area associated with this context.
       */
      public WorkArea obtain() 
      {
         return (WorkArea) this.get();
      }   
      
      /**
       * Initializes the work area, the first time it is requested within a
       * new context.
       *
       * @return   The newly instantiated work area.
       */
      protected synchronized Object initialValue()
      {
         return new WorkArea();
      }   
   }   
}
